import unittest

import numpy as np
from pytest import approx
import os
import pandas as pd

from Elasticipy.tensors.elasticity import StiffnessTensor, ComplianceTensor
from scipy.spatial.transform import Rotation
from Elasticipy.tensors.elasticity import _indices2str
from Elasticipy.crystal_symmetries import SYMMETRIES
from Elasticipy.tensors.stress_strain import StressTensor, StrainTensor
from pymatgen.analysis.elasticity import elastic as mg
from orix.quaternion import Rotation as orix_rot

current_dir = os.path.dirname(os.path.abspath(__file__))
file_path = os.path.join(current_dir, 'MaterialsProject.json')
data_base = pd.read_json(file_path)
rotations = Rotation.random(10000)

def variant_selection(symmetry, variant_name):
    for variant_group in symmetry.keys():
        elements = [elem.strip() for elem in variant_group.split(",")]
        if variant_name in elements:
            return symmetry[variant_group]
    return None

Smat = np.array([[8, -3, -2, 0, 14, 0],
                 [-3, 8, -5, 0, -8, 0],
                 [-2, -5, 10, 0, 0, 0],
                 [0, 0, 0, 12, 0, 0],
                 [14, -8, 0, 0, 116, 0],
                 [0, 0, 0, 0, 0, 12]])/1000
S = ComplianceTensor(Smat)


def crystal_symmetry_tester(symmetry_name, cls='stiffness', variant=None):
    symmetry = SYMMETRIES[symmetry_name]
    if variant is None:
        materials_of_interest = data_base[data_base.symmetry == symmetry_name]
        required_fields = symmetry.required
    else:
        materials_of_interest = data_base[data_base.point_group == variant]
        variant = variant_selection(symmetry, variant)
        required_fields = variant.required
    for index, row in materials_of_interest.iterrows():
        matrix = np.array(row['C'])
        if cls=='stiffness':
            class_constructor = StiffnessTensor
            prefix = 'C'
        else:
            class_constructor = ComplianceTensor
            prefix = 'S'
            matrix = np.linalg.inv(matrix)*1000
        kwargs = dict()
        for indices in required_fields:
            component_name = prefix + _indices2str(indices)
            kwargs[component_name] = matrix[tuple(indices)]
        constructor = getattr(class_constructor, symmetry_name.lower())
        C = constructor(**kwargs)
        assert np.all(C._matrix == approx(matrix, rel=0.5))
        return C


class TestComplianceTensor(unittest.TestCase):
    def test_young_modulus_eval(self):
        E = S.Young_modulus
        E_xyz = E.eval(np.eye(3))
        for i in range(3):
            self.assertEqual(E_xyz[i], 1/Smat[i, i])

    def test_young_modulus_stats(self):
        E = S.Young_modulus
        assert E.mean(method='exact') == approx(101.994)
        assert E.std() == approx(48.48264174566468)
        assert E.mean(method='trapezoid') == approx(101.9855, rel=1e-3)

    def test_shear_modulus_eval(self):
        G = S.shear_modulus
        u = [[0, 1, 0], [1, 0, 0], [1, 0, 0]]
        v = [[0, 0, 1], [0, 0, 1], [0, 1, 0]]
        G_xyz = G.eval(u, v)
        for i in range(3):
            self.assertEqual(G_xyz[i],  1/Smat[i+3, i+3])

    def test_Poisson_ratio_eval(self):
        nu = S.Poisson_ratio
        u = [[0, 1, 0], [1, 0, 0], [1, 0, 0]]
        v = [[0, 0, 1], [0, 0, 1], [0, 1, 0]]
        nu_xyz = nu.eval(u, v)
        nu_xyz_th = [0.625, 0.25, 0.375]
        for i in range(3):
            self.assertEqual(nu_xyz[i],  nu_xyz_th[i])

    def test_shear_modulus_mini_maxi(self):
        G = S.shear_modulus
        G_min, _ = G.min()
        G_max, _ = G.max()
        assert G_min == approx(8.47165)
        assert G_max == approx(83.3333)

    def test_unvoigt(self):
        lame1, lame2 = 1, 2
        C = StiffnessTensor.isotropic(lame1=lame1, lame2=lame2)
        C_full = C.full_tensor()
        eye = np.eye(3)
        A = np.einsum('ij,kl->ijkl', eye, eye)
        C_th = lame1 * A + 2 * lame2 * StiffnessTensor.identity(return_full_tensor=True)
        np.testing.assert_almost_equal(C_th, C_full)

    def test_averages(self):
        averages = [S.Voigt_average(), S.Reuss_average(), S.Hill_average()]
        E_mean_th = [151.738, 75.76, 114.45]
        G_mean_th = [55.653, 26.596, 41.124]
        nu_mean_th = [0.36325, 0.42424242, 0.3915]
        for i, average in enumerate(averages):
            assert approx(average.Young_modulus.eval([1,0,0]), rel=1e-4) == E_mean_th[i]
            assert approx(average.shear_modulus.eval([1,0,0],[0,1,0]), rel=1e-4) == G_mean_th[i]
            assert approx(average.Poisson_ratio.eval([1,0,0],[0,1,0]), rel=1e-4) == nu_mean_th[i]

    def test_wave_velocity(self, E=210, nu=0.3, rho=7.8):
        C = StiffnessTensor.isotropic(E=E, nu=nu)
        M = E * (1 - nu) / ((1 + nu) * (1 - 2 * nu))
        cp, cs_1, cs_2 = C.wave_velocity(rho)
        assert approx(cp.mean()) == np.sqrt(M / rho)
        G = C.shear_modulus.mean()
        assert approx(cs_2.mean()) == np.sqrt(G / rho)
        assert approx(cs_1.mean()) == np.sqrt(G / rho)

    def test_symmetry(self):
        S = np.random.random((6, 6))
        with self.assertRaises(ValueError) as context:
            _ = ComplianceTensor(S)
        self.assertEqual(str(context.exception), 'The input matrix must be symmetric')

    def test_positive_definite(self):
        S = np.array([
            [2, -1, 0, 0, 0, 0],
            [-1, 2, -1, 0, 0, 0],
            [0, -1, 2, -1, 0, 0],
            [0, 0, -1, 2, -1, 0],
            [0, 0, 0, -1, 2, -1],
            [0, 0, 0, 0, -1, 0]
        ])
        with self.assertRaises(ValueError) as context:
            _ = ComplianceTensor(S)
        eig_vals = np.linalg.eigvals(S)
        expected_error = 'The input matrix is not definite positive (eigenvalues: {})'.format(eig_vals)
        self.assertEqual(str(context.exception), expected_error)

    def test_input_matrix(self):
        for n in range(1,8):
            m = np.ones(shape=(n,n))
            if n != 6:
                with self.assertRaises(ValueError) as context:
                    _ = ComplianceTensor(m)
                self.assertEqual(str(context.exception), 'The input matrix must of shape (...,6,6) or (...,3,3,3,3)')

    def test_full_tensor_as_input(self):
        a = ComplianceTensor.isotropic(E=210, nu=0.3)
        b = ComplianceTensor(a.full_tensor())
        np.testing.assert_array_almost_equal(a._matrix, b._matrix)


    def test_component(self):
        assert S.C11 == Smat[0, 0]
        assert S.C12 == Smat[0, 1]
        assert S.C13 == Smat[0, 2]
        assert S.C14 == Smat[0, 3]
        assert S.C15 == Smat[0, 4]
        assert S.C16 == Smat[0, 5]
        assert S.C22 == Smat[1, 1]
        assert S.C23 == Smat[1, 2]
        assert S.C24 == Smat[1, 3]
        assert S.C25 == Smat[1, 4]
        assert S.C26 == Smat[1, 5]
        assert S.C32 == Smat[2, 1]
        assert S.C33 == Smat[2, 2]
        assert S.C34 == Smat[2, 3]
        assert S.C35 == Smat[2, 4]
        assert S.C36 == Smat[2, 5]
        assert S.C44 == Smat[3, 3]
        assert S.C45 == Smat[3, 4]
        assert S.C46 == Smat[3, 5]
        assert S.C55 == Smat[4, 4]
        assert S.C56 == Smat[4, 5]
        assert S.C66 == Smat[5, 5]
        docstring = ComplianceTensor.C12.__doc__
        assert docstring == 'Returns the (1,2) component of the Compliance matrix.'

    def test_bulk_modulus(self):
        E, nu = 210, 0.3
        Siso= ComplianceTensor.isotropic(E=E, nu=nu)
        assert Siso.bulk_modulus == approx(E / (3 * (1-2 * nu)))

    def test_mean(self):
        m, n = 5, 10
        orix_rotations = orix_rot.random((m, n))
        S_rotated = S * orix_rotations
        S_mean_0 = S_rotated.mean(axis=0)
        for i in range(n):
            np.testing.assert_array_almost_equal(S_mean_0[i]._matrix, S_rotated[:, i].Reuss_average()._matrix)
        S_mean_1 = S_rotated.mean(axis=1)
        for i in range(m):
            np.testing.assert_array_almost_equal(S_mean_1[i]._matrix, S_rotated[i, :].Reuss_average()._matrix)
        S_mean = S_rotated.mean()
        np.testing.assert_array_almost_equal(S_mean._matrix, S_rotated.flatten().Reuss_average()._matrix)
        np.testing.assert_array_almost_equal(S_mean._matrix, S_rotated.Reuss_average()._matrix)

    def test_to_from_Kelvin(self):
        matrix = S.to_Kelvin()
        S2 = ComplianceTensor.from_Kelvin(matrix, symmetry=S.symmetry)
        np.testing.assert_array_almost_equal(S._matrix, S2._matrix)

    def test_full_tensor(self):
        S_full = S.full_tensor()
        S2 = ComplianceTensor(S_full)
        assert S == S2
        np.testing.assert_array_equal(S.full_tensor(), S2.full_tensor())

    def test_repr(self):
        """Test printing out the tensor"""
        str = S.__repr__()
        assert str == ('Compliance tensor (in Voigt mapping):\n'
                       '[[ 0.008 -0.003 -0.002  0.     0.014  0.   ]\n'
                       ' [-0.003  0.008 -0.005  0.    -0.008  0.   ]\n'
                       ' [-0.002 -0.005  0.01   0.     0.     0.   ]\n'
                       ' [ 0.     0.     0.     0.012  0.     0.   ]\n'
                       ' [ 0.014 -0.008  0.     0.     0.116  0.   ]\n'
                       ' [ 0.     0.     0.     0.     0.     0.012]]'
                       '\nSymmetry: Triclinic')

        S_rotated = S * rotations
        assert S_rotated.__repr__() ==  'Compliance tensor array of shape ({},)\nSymmetry: Triclinic'.format(len(rotations))

    def test_multidimensional_attributes(self):
        S_rotated = S * rotations
        expected = 'Young_modulus is not suitable for tensor array. Consider subscripting (e.g. C[0].Young_modulus).'
        with self.assertRaises(ValueError) as context:
            _ = S_rotated.Young_modulus
        self.assertEqual(str(context.exception), expected)

        expected = 'shear_modulus is not suitable for tensor array. Consider subscripting (e.g. C[0].shear_modulus).'
        with self.assertRaises(ValueError) as context:
            _ = S_rotated.shear_modulus
        self.assertEqual(str(context.exception), expected)

        expected = 'linear_compressibility is not suitable for tensor array. Consider subscripting (e.g. C[0].linear_compressibility).'
        with self.assertRaises(ValueError) as context:
            _ = S_rotated.linear_compressibility
        self.assertEqual(str(context.exception), expected)

        expected = 'wave_velocity is not suitable for tensor array. Consider subscripting (e.g. C[0].wave_velocity).'
        with self.assertRaises(ValueError) as context:
            _ = S_rotated.wave_velocity(5)
        self.assertEqual(str(context.exception), expected)

    def test_voigt_reuss_axis(self):
        m,n = 5,6
        orientations = orix_rot.random((m,n))
        S_rotated = S * orientations
        Sv_0 = S_rotated.Voigt_average(axis=0)
        Sr_0 = S_rotated.Reuss_average(axis=0)
        assert Sv_0.shape == (n,)
        assert Sr_0.shape == (n,)
        for i in range(n):
            assert Sv_0[i] == S_rotated[:, i].Voigt_average()
            assert Sr_0[i] == S_rotated[:, i].Reuss_average()
        Sv_1 = S_rotated.Voigt_average(axis=1)
        Sr_1 = S_rotated.Reuss_average(axis=1)
        assert Sv_1.shape == (m,)
        assert Sr_1.shape == (m,)
        for i in range(m):
            assert Sv_1[i] == S_rotated[i, :].Voigt_average()
            assert Sr_1[i] == S_rotated[i, :].Reuss_average()

        np.testing.assert_array_almost_equal(Sv_0.Voigt_average()._matrix, S_rotated.Voigt_average()._matrix)
        np.testing.assert_array_almost_equal(Sv_1.Voigt_average()._matrix, S_rotated.Voigt_average()._matrix)
        np.testing.assert_array_almost_equal(Sr_0.Reuss_average()._matrix, S_rotated.Reuss_average()._matrix)
        np.testing.assert_array_almost_equal(Sr_1.Reuss_average()._matrix, S_rotated.Reuss_average()._matrix)

    def test_invariants(self):
        inv = S.invariants()
        S_rotated = S * rotations
        for i, inv_rotated_i in enumerate(S_rotated.invariants()):
            np.testing.assert_array_almost_equal(inv_rotated_i, inv[i])

    def test_deprecated_path(self):
        expected_warn = ("The module 'Elasticipy.FourthOrderTensor' is deprecated and will be removed in a future "
                            "release. Please use 'Elasticipy.tensors.elasticity' instead.")
        with self.assertWarns(DeprecationWarning) as context:
            from Elasticipy.FourthOrderTensor import ComplianceTensor, StiffnessTensor
        self.assertEqual(str(context.warning), expected_warn)


class TestStiffnessConstructor(unittest.TestCase):
    def test_averages(self):
        """Check that the Voigt, Reuss and Hill averages are consistent with those provided by MP."""
        rel = 5e-2
        for index, row in data_base.iterrows():
            matrix = row['C']
            symmetry = row['symmetry']
            C = StiffnessTensor(matrix, symmetry=symmetry)
            C_rotated = C * rotations
            for method in ('voigt', 'reuss', 'hill', 'dummy'):
                if method == 'dummy':
                    with self.assertRaises(NotImplementedError) as context:
                        _ = C_rotated.average(method)
                    self.assertEqual(str(context.exception), 'Only Voigt, Reus, and Hill are implemented.')
                else:
                    Gavg = C.average(method).shear_modulus.mean(n_evals=10000)
                    assert row['G' + method] == approx(Gavg, rel=rel)
                    Gavg = C_rotated.average(method).shear_modulus.mean(n_evals=10000)
                    assert row['G' + method] == approx(Gavg, rel=rel)

    def test_isotropic(self):
        E, nu = 210000, 0.28
        C = StiffnessTensor.isotropic(E=E, nu=nu)
        G = C.shear_modulus.mean()
        K = C.bulk_modulus
        lame1 = C.lame1
        lame2 = C.lame2
        assert approx(G) == E / (1+nu) /2
        input_args1 = {'E': E, 'nu': nu, 'K':K, 'lame1':lame1, 'lame2':lame2}
        input_args2 = {'E': E, 'nu': nu, 'K':K, 'lame1':lame1, 'lame2':lame2}
        for k1 in input_args1.keys():
            del input_args2[k1]
            for k2 in input_args2.keys():
                kwargs = {k1: input_args1[k1], k2: input_args2[k2]}
                C_combo = StiffnessTensor.isotropic(**kwargs)
                np.testing.assert_array_almost_equal(C.matrix(), C_combo.matrix())

        with self.assertRaises(ValueError) as context:
            _ = StiffnessTensor.isotropic(E=E, nu=nu, G=G)
        self.assertEqual(str(context.exception), "Exactly two values are required among E, nu, G, K, lame1 and lame2.")
        with self.assertRaises(ValueError) as context:
            _ = StiffnessTensor.isotropic(G=G, lame2=G)
        self.assertEqual(str(context.exception), "G and lame2 cannot be provided together.")
        assert approx(C.Young_modulus.mean()) == E
        assert C.is_isotropic()
        assert C.is_cubic()
        assert C.is_tetragonal()

    def test_stiffness_cubic(self):
        """Check that all symmetries in stiffness are well taken into account for cubic case"""
        C = crystal_symmetry_tester('Cubic')
        shape = (50,100)
        C_rotated = C * (orix_rot.random(shape))
        is_cubic = C_rotated.is_cubic()
        assert is_cubic.shape == shape
        assert np.all(is_cubic)
        assert np.all(C.is_tetragonal())
        assert not np.any(C.is_isotropic())

    def test_stiffness_hexagonal(self):
        """Check that all symmetries in stiffness are well taken into account for hexagonal case"""
        C = crystal_symmetry_tester('Hexagonal')
        assert not C.is_tetragonal()
        assert not np.any(C.is_cubic())

    def test_stiffness_trigonal(self):
        """Check that all symmetries in stiffness are well taken into account for trigonal case"""
        C = crystal_symmetry_tester('Trigonal', variant='32')
        C_rotated = C * rotations
        assert not np.any(C_rotated.is_tetragonal())
        C = crystal_symmetry_tester('Trigonal', variant='-3')
        C_rotated = C * rotations
        assert not np.any(C_rotated.is_tetragonal())

    def test_stiffness_tetragonal(self):
        """Check that all symmetries in stiffness are well taken into account for tetragonal case"""
        C = crystal_symmetry_tester('Tetragonal', variant='-42m')
        C_rotated = C * rotations
        assert np.all(C_rotated.is_tetragonal())
        C = crystal_symmetry_tester('Tetragonal', variant='-4')
        C_rotated = C * rotations
        assert np.all(C_rotated.is_tetragonal())

    def test_stiffness_orthorhombic(self):
        """Check that all symmetries in stiffness are well taken into account for orthorhombic case"""
        crystal_symmetry_tester('Orthorhombic')

    def test_stiffness_monoclinic(self):
        """Check that all symmetries in stiffness are well taken into account for monoclinic case"""
        crystal_symmetry_tester('Monoclinic', variant='Diad || y')

    def test_compliance_cubic(self):
        """Check that all symmetries in compliance are well taken into account for cubic case"""
        crystal_symmetry_tester('Cubic', cls='compliance')

    def test_compliance_hexagonal(self):
        """Check that all symmetries in compliance are well taken into account for hexagonal case"""
        crystal_symmetry_tester('Hexagonal', cls='compliance')

    def test_compliance_trigonal(self):
        """Check that all symmetries in compliance are well taken into account for trigonal case"""
        crystal_symmetry_tester('Trigonal', variant='32', cls='compliance')
        crystal_symmetry_tester('Trigonal', variant='-3', cls='compliance')

    def test_compliance_tetragonal(self):
        """Check that all symmetries in compliance are well taken into account for tetragonal case"""
        crystal_symmetry_tester('Tetragonal', variant='-42m', cls='compliance')
        crystal_symmetry_tester('Tetragonal', variant='-4', cls='compliance')

    def test_compliance_orthorhombic(self):
        """Check that all symmetries in compliance are well taken into account for orthorhombic case"""
        crystal_symmetry_tester('Orthorhombic', cls='compliance')

    def test_compliance_monoclinic(self):
        """Check that all symmetries in compliance are well taken into account for monoclinic case"""
        crystal_symmetry_tester('Monoclinic', variant='Diad || y', cls='compliance')

    def test_young_modulus_eval(self):
        """Check that the Young modulus is given somehow given by the compliance tensor"""
        E = S.Young_modulus
        E_xyz = E.eval(np.eye(3))
        for i in range(3):
            self.assertEqual(E_xyz[i], 1/Smat[i, i])

    def test_young_modulus_stats(self):
        """Test statistics for Young moduli"""
        E = S.Young_modulus
        assert E.mean(method='exact') == approx(101.994)
        assert E.std() == approx(48.48264174566468)
        assert E.mean() == approx(101.9942123)

    def test_shear_modulus_eval(self):
        """Test shear moduli estimations"""
        G = S.shear_modulus
        u = [[0, 1, 0], [1, 0, 0], [1, 0, 0]]
        v = [[0, 0, 1], [0, 0, 1], [0, 1, 0]]
        G_xyz = G.eval(u, v)
        for i in range(3):
            self.assertEqual(G_xyz[i],  1/Smat[i+3, i+3])

    def test_Poisson_ratio_eval(self):
        """Test Poisson ration estimations"""
        nu = S.Poisson_ratio
        u = [[0, 1, 0], [1, 0, 0], [1, 0, 0]]
        v = [[0, 0, 1], [0, 0, 1], [0, 1, 0]]
        nu_xyz = nu.eval(u, v)
        nu_xyz_th = [0.625, 0.25, 0.375]
        for i in range(3):
            self.assertEqual(nu_xyz[i],  nu_xyz_th[i])

    def test_shear_modulus_mini_maxi(self):
        """Test shear min/max"""
        G = S.shear_modulus
        G_min, _ = G.min()
        G_max, _ = G.max()
        assert G_min == approx(8.47165)
        assert G_max == approx(83.3333)

    def test_unvoigt(self):
        """Test if the isotropic second-order tensor is well reconstructed"""
        lame1, lame2 = 1, 2
        C = StiffnessTensor.isotropic(lame1=lame1, lame2=lame2)
        C_full = C.full_tensor()
        eye = np.eye(3)
        A = np.einsum('ij,kl->ijkl', eye, eye)
        B = np.einsum('ik,jl->ijkl', eye, eye)
        C = np.einsum('il,kj->ijkl', eye, eye)
        C_th = lame1 * A + lame2 * (B + C)
        np.testing.assert_almost_equal(C_th, C_full)

    def test_wave_velocity(self):
        """Test computation of wave velocities against simple isotropic case"""
        E = 210
        nu = 0.3
        rho = 7.8
        C = StiffnessTensor.isotropic(E=E, nu=nu)
        M = E * (1 - nu) / ((1 + nu) * (1 - 2 * nu))
        cp, cs_1, cs_2 = C.wave_velocity(rho)
        assert approx(cp.mean()) == np.sqrt(M / rho)
        G = C.shear_modulus.mean()
        assert approx(cs_2.mean()) == np.sqrt(G / rho)
        assert approx(cs_1.mean()) == np.sqrt(G / rho)

    def test_monoclinic(self):
        """Test constructor for monoclinic symmetry"""
        common_arguments = {'C11':11, 'C12':12, 'C13':13, 'C22':22, 'C23':23, 'C33':33, 'C44':44, 'C55':55, 'C66':66}

        # Check for Diad||y
        C = StiffnessTensor.monoclinic(**common_arguments, C16=16, C26=26, C36=36, C45=45)
        matrix = np.array([[11, 12, 13, 0, 0, 16],
                           [12, 22, 23, 0, 0, 26],
                           [13, 23, 33, 0, 0, 36],
                           [0,  0,  0, 44, 45, 0],
                           [0,  0,  0, 45, 55, 0],
                           [16, 26, 36, 0, 0, 66]], dtype=np.float64)
        np.testing.assert_array_equal(matrix, C._matrix)

        # Check for Diad||z
        C = StiffnessTensor.monoclinic(**common_arguments, C15=15, C25=25, C35=35, C46=46)
        matrix = np.array([[11, 12, 13, 0, 15, 0],
                           [12, 22, 23, 0,  25, 0],
                           [13, 23, 33, 0,  35, 0],
                           [0,  0,  0,  44, 0, 46],
                           [15, 25, 35, 0,  55, 0],
                           [0,  0,  0,  46, 0, 66]], dtype=np.float64)
        np.testing.assert_array_equal(matrix, C._matrix)

        # Check ambiguous cases
        expected_error = "'Ambiguous diad. Provide either C15, C25, C35 and C46; or C16, C26, C36 and C45'"
        with self.assertRaises(KeyError) as context:
            C = StiffnessTensor.monoclinic(**common_arguments,
                                          C15=15, C25=25, C35=35, C46=46, C16=16, C26=26, C36=36, C45=45)
        self.assertEqual(str(context.exception), expected_error)

        expected_error = ("'For monoclinic symmetry, one should provide either C15, C25, C35 and C46, "
                          "or C16, C26, C36 and C45.'")
        with self.assertRaises(KeyError) as context:
            C = StiffnessTensor.monoclinic(**common_arguments)
        self.assertEqual(str(context.exception), expected_error)

    def test_write_read_tensor(self):
        """Test export and import stiffness tensor to text file"""
        filename = 'C_tmp.txt'
        C = StiffnessTensor.isotropic(E=210, nu=0.3)
        C.save_to_txt(filename)
        C2 = StiffnessTensor.from_txt_file(filename)
        np.testing.assert_allclose(C2._matrix, C._matrix, atol=1e-2)

        # Now the same with phase name
        C = StiffnessTensor.isotropic(E=210, nu=0.3, phase_name='Steel')
        C.save_to_txt(filename)
        C2 = StiffnessTensor.from_txt_file(filename)
        np.testing.assert_allclose(C2._matrix, C._matrix, atol=1e-2)
        assert C2.phase_name == 'Steel'


    def test_equality(self):
        """Test == operator"""
        C1 = StiffnessTensor.isotropic(E=210000, nu=0.3)
        C2 = StiffnessTensor.isotropic(E=210000, nu=0.3)
        assert C1 == C2
        assert C1 == C2._matrix

    def test_add_sub(self):
        """Test addition and subtraction of tensors"""
        C1 = StiffnessTensor.isotropic(E=200, nu=0.3)
        C2 = StiffnessTensor.isotropic(E=100, nu=0.3)
        C_plus = C1 + C2
        assert C_plus.Young_modulus.mean() == approx(300)
        C_minus = C1 - C2
        assert C_minus.Young_modulus.mean() == approx(100)
        C_minus = C1 - C2._matrix
        assert C_minus.Young_modulus.mean() == approx(100)
        C_plus_full = C1 + C2.full_tensor()
        assert C_plus_full == C_plus

        with self.assertRaises(ValueError) as context:
            _ = C1 + C2.inv()
        self.assertEqual(str(context.exception), 'The two tensors to add must be of the same class.')

    def test_mul_rmul(self):
        C = StiffnessTensor.isotropic(E=200, nu=0.3)
        C1 = C * 2
        C2 = 2 * C
        assert C1 == C2

    def test_div(self):
        C = StiffnessTensor.isotropic(E=200, nu=0.3)
        Cdiv = C/2
        np.testing.assert_array_almost_equal(Cdiv._matrix, C._matrix / 2)

    def test_weighted_average(self):
        """Test averaging two phases"""
        E1 = 100
        E2 = 200
        C1 = StiffnessTensor.isotropic(E=E1, nu=0.3)
        C2 = StiffnessTensor.isotropic(E=E2, nu=0.3)
        Cv = StiffnessTensor.weighted_average((C1, C2), [0.5, 0.5], method='Voigt')
        Cr = StiffnessTensor.weighted_average((C1, C2), [0.5, 0.5], method='Reuss')
        Ch = StiffnessTensor.weighted_average((C1, C2), [0.5, 0.5], method='Hill')
        E_voigt = (E1 + E2) / 2
        E_reuss = 2 / (1/E1 + 1/E2)
        assert Cv.Young_modulus.mean() == approx(E_voigt)
        assert Cr.Young_modulus.mean() == approx(E_reuss)
        assert Ch.Young_modulus.mean() == approx(E_voigt/2 + E_reuss/2)

    def test_orthotropic(self):
        """Check if the engineering constants are well retrieved in the orthotropic case"""
        Ex, Ey, Ez = 100., 200., 300.
        nu_yx, nu_zy, nu_zx = 0.2, 0.3, 0.4
        G_xy, G_xz, G_yz = 50., 60., 70.
        C = StiffnessTensor.orthotropic(Ex=Ex, Ey=Ey, Ez=Ez, nu_yx=nu_yx, nu_zx=nu_zx, nu_zy=nu_zy,
                                        Gxy=G_xy, Gxz=G_xz, Gyz=G_yz)
        E = C.Young_modulus
        assert E.eval([1,0,0]) == approx(Ex)
        assert E.eval([0,1,0]) == approx(Ey)
        assert E.eval([0,0,1]) == approx(Ez)
        G = C.shear_modulus
        assert G.eval([1,0,0], [0,1,0]) == approx(G_xy)
        assert G.eval([1, 0, 0], [0, 0, 1]) == approx(G_xz)
        assert G.eval([0, 1, 0], [0, 0, 1]) == approx(G_yz)
        nu = C.Poisson_ratio
        assert nu.eval([0,1,0], [1,0,0]) == approx(nu_yx)
        assert nu.eval([0, 0, 1], [0, 1, 0]) == approx(nu_zy)
        assert nu.eval([0, 0, 1], [1, 0, 0]) == approx(nu_zx)

        # Now check with "inverted" Poisson ratio
        nu_xy = nu.eval([1, 0, 0], [0, 1, 0])
        nu_xz = nu.eval([1, 0, 0], [0, 0, 1])
        nu_yz = nu.eval([0, 1, 0], [0, 0, 1])
        C2 = StiffnessTensor.orthotropic(Ex=Ex, Ey=Ey, Ez=Ez, nu_xy=nu_xy, nu_xz=nu_xz, nu_yz=nu_yz,
                                        Gxy=G_xy, Gxz=G_xz, Gyz=G_yz)
        np.testing.assert_array_almost_equal(C._matrix, C2._matrix)

        # Now try passing both nu_yz and nu_zy
        with self.assertRaises(ValueError) as context:
            StiffnessTensor.orthotropic(Ex=Ex, Ey=Ey, Ez=Ez, nu_xy=nu_xy, nu_xz=nu_xz, nu_yz=nu_yz, nu_zy=nu_zy,
                                        Gxy=G_xy, Gxz=G_xz, Gyz=G_yz)
        self.assertEqual(str(context.exception), 'Either nu_yz or nu_zy must be provided')

        # Check that the Poisson ratios are consistent with tensile test
        stress = StressTensor.tensile([1,0,0],1)
        strain = C.inv()*stress
        assert strain.C[1,1]/strain.C[0,0] == approx(-nu_xy)
        assert strain.C[2,2]/strain.C[0,0] == approx(-nu_xz)
        stress = StressTensor.tensile([0,1,0],1)
        strain = C.inv()*stress
        assert strain.C[0,0]/strain.C[1,1] == approx(-nu_yx)
        assert strain.C[2,2]/strain.C[1,1] == approx(-nu_yz)

    def test_transverse_isotropic(self):
        """Check if the engineering constants are well retrieved in the transverse-isotropic case"""
        Ex, Ez = 100., 200.
        nu_yx, nu_zx = 0.2, 0.3
        Gxz = 80
        C = StiffnessTensor.transverse_isotropic(Ex=Ex, Ez=Ez, nu_yx=nu_yx, nu_zx=nu_zx, Gxz=Gxz)
        E = C.Young_modulus
        assert E.eval([1,0,0]) == approx(Ex)
        assert E.eval([0,1,0]) == approx(Ex)
        assert E.eval([0,0,1]) == approx(Ez)
        G = C.shear_modulus
        assert G.eval([1, 0, 0], [0, 0, 1]) == approx(Gxz)
        nu = C.Poisson_ratio
        assert nu.eval([0, 1, 0], [1, 0, 0]) == approx(nu_yx)
        assert nu.eval([0, 0, 1], [0, 1, 0]) == approx(nu_zx)
        assert nu.eval([0, 0, 1], [1, 0, 0]) == approx(nu_zx)

        # Try passing the opposite Poisson ratios
        nu_xy = nu.eval([1, 0, 0], [0, 1, 0])
        nu_xz = nu.eval([1, 0, 0], [0, 0, 1])
        C2 = StiffnessTensor.transverse_isotropic(Ex=Ex, Ez=Ez, nu_xy=nu_xy, nu_xz=nu_xz, Gxz=Gxz)
        np.testing.assert_array_almost_equal(C._matrix, C2._matrix)

        # Now try passing both nu_xy and nu_yx
        with self.assertRaises(ValueError) as context:
            StiffnessTensor.transverse_isotropic(Ex=Ex, Ez=Ez, nu_yx=nu_yx, nu_zx=nu_zx, Gxz=Gxz, nu_xy=nu_xy)
        self.assertEqual(str(context.exception), 'Either nu_xy or nu_yx must be provided')

    def test_straining_energy(self):
        """Test if the elastic energies are consistent."""
        matrix = np.random.random((3,3))
        stress = StressTensor(matrix + matrix.T)
        strain = S * stress
        e1 = stress.elastic_energy(strain)
        e2 = strain.elastic_energy(stress)
        np.testing.assert_approx_equal(e1, e2)


    def test_to_pymatgen(self):
        """Test exporting stiffness and compliance to pymatgen format"""
        C = S.inv()
        Cvrh = C.Hill_average()
        Cvrh_pymatgen = mg.ElasticTensor(Cvrh.full_tensor())
        C_pymatgen = mg.ElasticTensor(C.full_tensor())
        assert Cvrh_pymatgen.y_mod == approx(Cvrh.Young_modulus.mean()*1e9)
        assert C_pymatgen.g_vrh == approx(Cvrh.shear_modulus.mean())

        S_pymatgen = S.to_pymatgen()
        np.testing.assert_array_almost_equal(S_pymatgen.voigt, np.linalg.inv(C_pymatgen.voigt))

    def test_MaterialsProject(self):
        """Test import from the Materials Project"""
        # Try with cubic Cu
        C = StiffnessTensor.from_MP("mp-30")
        C_Cu = StiffnessTensor.cubic(C11=186, C12=134, C44=77)
        np.testing.assert_array_almost_equal(C._matrix, C_Cu._matrix)

        # Now try with a list of entries
        Cs = StiffnessTensor.from_MP(("mp-30", "mp-1048"))
        assert len(Cs) == 2
        np.testing.assert_array_almost_equal(Cs[0]._matrix, C_Cu._matrix)


    def test_getitem(self):
        """Test indexing of stiffness tensor"""
        S_rotated = S * rotations
        S1 = S_rotated[0]
        S2 = S * rotations[0]
        np.testing.assert_array_almost_equal(S1._matrix, S2._matrix)
        expected_error = 'A single tensor cannot be subindexed'
        with self.assertRaises(IndexError) as context:
            _ = S[0]
        self.assertEqual(str(context.exception), expected_error)

    def test_symmetry(self):
        S = np.random.random((6, 6))
        with self.assertRaises(ValueError) as context:
            _ = StiffnessTensor(S)
        self.assertEqual(str(context.exception), 'The input matrix must be symmetric')

    def test_positive_definite(self):
        S = np.array([
            [2, -1, 0, 0, 0, 0],
            [-1, 2, -1, 0, 0, 0],
            [0, -1, 2, -1, 0, 0],
            [0, 0, -1, 2, -1, 0],
            [0, 0, 0, -1, 2, -1],
            [0, 0, 0, 0, -1, 0]
        ])
        with self.assertRaises(ValueError) as context:
            _ = StiffnessTensor(S)
        eig_vals = np.linalg.eigvals(S)
        expected_error = 'The input matrix is not definite positive (eigenvalues: {})'.format(eig_vals)
        self.assertEqual(str(context.exception), expected_error)

    def test_Zener_universal_anisotropy(self):
        C11, C12, C44 = 173, 33, 18
        C = StiffnessTensor.cubic(C11=C11, C12=C12, C44=C44)
        Z = C.Zener_ratio()
        assert 6/5 * (Z**0.5 - Z**(-0.5))**2 == approx(C.universal_anisotropy)
        C_cub_iso = StiffnessTensor.cubic(C11=C11, C12=C12, C44=(C11-C12)/2)
        assert C_cub_iso.Zener_ratio() == 1.0
        Ciso = StiffnessTensor.isotropic(E=210, nu=0.3)
        assert Ciso.Zener_ratio() == 1.0
        assert Ciso.universal_anisotropy < 1e-10
        Cmono = S.inv()
        with self.assertRaises(ValueError) as context:
            Cmono.Zener_ratio()
        self.assertEqual(str(context.exception), 'The tensor does not seem to have cubic symmetry within the given tolerance (0.0001)')
        C_rot = C*rotations[0]
        Zrot = C_rot.Zener_ratio()
        assert Zrot == approx(Z)

    def test_orix(self):
        # Orix allows multidimensional arrays of rotations
        m, n= 5,6
        orix_rotations = orix_rot.random((m,n))
        C = S.inv()
        C_rotated = C * orix_rotations
        C_rotated_full = C_rotated.full_tensor()
        for i in range(m):
            for j in range(n):
                inv_rotation = ~orix_rotations
                rot_mat = inv_rotation.to_matrix()[i,j]
                tensor_i = np.einsum('im,jn,ko,lp,mnop -> ijkl', rot_mat, rot_mat, rot_mat, rot_mat, C.full_tensor())
                np.testing.assert_array_almost_equal(C_rotated_full[i,j], tensor_i)

        # Check that the result is consistent with scipy.Rotation
        orix_rotations = orix_rotations.transpose().flatten()   # orix seems to use the opposite order from numpy for flattening
        C_rotated_flat = C_rotated.flatten()
        euler = orix_rotations.to_euler()
        scipy_rotations = Rotation.from_euler('ZXZ', euler)
        C_rotated_scipy = C * scipy_rotations
        C_rotated_full_scipy = C_rotated_scipy.full_tensor()
        np.testing.assert_array_almost_equal(C_rotated_full_scipy, C_rotated_flat.full_tensor())


    def test_linear_compressibility(self):
        E, nu = 210, 0.3
        Ciso = StiffnessTensor.isotropic(E=E, nu=nu)
        beta = Ciso.linear_compressibility.mean()
        assert Ciso.bulk_modulus == approx(1 / (3*beta))

    def test_full_tensor_as_input(self):
        a = StiffnessTensor.isotropic(E=210, nu=0.3)
        b = StiffnessTensor(a.full_tensor())
        assert a == b

    def test_component(self):
        C = S.inv()
        Cmat = C._matrix
        assert C.C11 == Cmat[0, 0]
        assert C.C12 == Cmat[0, 1]
        assert C.C13 == Cmat[0, 2]
        assert C.C14 == Cmat[0, 3]
        assert C.C15 == Cmat[0, 4]
        assert C.C16 == Cmat[0, 5]
        assert C.C22 == Cmat[1, 1]
        assert C.C23 == Cmat[1, 2]
        assert C.C24 == Cmat[1, 3]
        assert C.C25 == Cmat[1, 4]
        assert C.C26 == Cmat[1, 5]
        assert C.C32 == Cmat[2, 1]
        assert C.C33 == Cmat[2, 2]
        assert C.C34 == Cmat[2, 3]
        assert C.C35 == Cmat[2, 4]
        assert C.C36 == Cmat[2, 5]
        assert C.C44 == Cmat[3, 3]
        assert C.C45 == Cmat[3, 4]
        assert C.C46 == Cmat[3, 5]
        assert C.C55 == Cmat[4, 4]
        assert C.C56 == Cmat[4, 5]
        assert C.C66 == Cmat[5, 5]
        docstring = StiffnessTensor.C12.__doc__
        assert docstring == 'Returns the (1,2) component of the Stiffness matrix.'

    def test_bulk_modulus(self):
        E, nu = 210, 0.3
        Ciso= StiffnessTensor.isotropic(E=E, nu=nu)
        assert Ciso.bulk_modulus == approx(E / (3 * (1-2 * nu)))

        b = S.bulk_modulus
        C = S.inv()
        C_rotated = C * rotations
        b_r = C_rotated.bulk_modulus
        assert np.allclose(b, b_r)

    def test_broadcasting(self):
        C11, C12, C44 = 173, 33, 18
        m,n,o = 9,7,5
        rot_0d = orix_rot.random()
        rot_1d = orix_rot.random(o)
        rot_2d = orix_rot.random((n, o))
        rot_3d = orix_rot.random((m, n, o))
        C_0d = StiffnessTensor.cubic(C11=C11, C12=C12, C44=C44)
        C_rotated_0d = C_0d * rot_0d
        C_rotated_1d = C_0d * rot_1d
        C_rotated_2d = C_0d * rot_2d
        C_rotated_3d = C_0d * rot_3d
        strain_0d = StrainTensor.rand()
        strain_1d = StrainTensor.rand((o,))
        strain_2d = StrainTensor.rand((n, o))
        strain_3d = StrainTensor.rand((m, n, o))

        stress = C_rotated_0d * strain_1d
        for i in range(o):
            assert stress[i] == C_rotated_0d * strain_1d[i]
        stress = C_rotated_1d * strain_0d
        for i in range(o):
            np.testing.assert_array_almost_equal(stress[i].matrix, (C_rotated_1d[i] * strain_0d).matrix)
        stress = C_rotated_1d * strain_1d
        for i in range(o):
            np.testing.assert_array_almost_equal(stress[i].matrix, (C_rotated_1d[i] * strain_1d[i]).matrix)
        stress = C_rotated_2d * strain_0d
        for i in range(n):
            for j in range(o):
                np.testing.assert_array_almost_equal(stress[i, j].matrix, (C_rotated_2d[i, j] * strain_0d).matrix)
        stress = C_rotated_2d * strain_1d
        for i in range(n):
            for j in range(o):
                np.testing.assert_array_almost_equal(stress[i,j].matrix, (C_rotated_2d[i,j] * strain_1d[j]).matrix)
        stress = C_rotated_2d * strain_2d
        for i in range(n):
            for j in range(o):
                np.testing.assert_array_almost_equal(stress[i,j].matrix, (C_rotated_2d[i,j] * strain_2d[i,j]).matrix)
        stress = C_rotated_3d * strain_0d
        for i in range(m):
            for j in range(n):
                for k in range(o):
                    np.testing.assert_array_almost_equal(stress[i,j,k].matrix, (C_rotated_3d[i,j,k] * strain_0d).matrix)
        stress = C_rotated_3d * strain_1d
        for i in range(m):
            for j in range(n):
                for k in range(o):
                    np.testing.assert_array_almost_equal(stress[i, j, k].matrix,
                                                         (C_rotated_3d[i, j, k] * strain_1d[k]).matrix)
        stress = C_rotated_3d * strain_2d
        for i in range(m):
            for j in range(n):
                for k in range(o):
                    np.testing.assert_array_almost_equal(stress[i, j, k].matrix,
                                                         (C_rotated_3d[i, j, k] * strain_2d[j,k]).matrix)
        stress = C_rotated_3d * strain_3d
        for i in range(m):
            for j in range(n):
                for k in range(o):
                    np.testing.assert_array_almost_equal(stress[i, j, k].matrix,
                                                         (C_rotated_3d[i, j, k] * strain_3d[i,j,k]).matrix)

    def test_transpose_array(self):
        m, n, o = 5, 6, 7
        rot_2d = orix_rot.random((m, n, o))
        C11, C12, C44 = 173, 33, 18
        C = StiffnessTensor.cubic(C11=C11, C12=C12, C44=C44)
        C_rotated = C * rot_2d
        C_rotated_T = C_rotated.transpose_array()
        assert C_rotated_T.shape == (o, n, m)
        for i in range(m):
            for j in range(n):
                for k in range(o):
                    np.testing.assert_array_almost_equal(C_rotated_T[k,j,i].full_tensor(), C_rotated[i,j,k].full_tensor())


    def test_eigenstiffness_eigencompliance(self):
        C11, C12, C44 = 22, 12, 44
        C = StiffnessTensor.cubic(C11=C11, C12=C12, C44=C44)
        eigen_stiffnesses = C.eig_stiffnesses
        eigen_stiffnesses_th = [C11 + 2 * C12, C11 - C12, 2 * C44] # 10.1111/j.1365-2478.2011.01049.x
        for e in eigen_stiffnesses:
            assert np.any(np.isclose(e, eigen_stiffnesses_th))
        eigen_strains = C.eig_strains
        for i, e in enumerate(eigen_strains.T):
            strain = StrainTensor.from_Kelvin(e)
            stress = C * strain
            np.testing.assert_array_almost_equal(stress.matrix, strain.matrix * eigen_stiffnesses[i])

        S = C.inv()
        eigen_compliances = S.eig_compliances
        eigen_compliances_th = 1 / np.array([C11 + 2 * C12, C11 - C12, 2 * C44])  # 10.1111/j.1365-2478.2011.01049.x
        for e in eigen_compliances:
            assert np.any(np.isclose(e, eigen_compliances_th))
        eigen_stresses = S.eig_stresses
        for i, e in enumerate(eigen_stresses.T):
            stress = StressTensor.from_Kelvin(e)
            strain = S * stress
            np.testing.assert_array_almost_equal(strain.matrix, stress.matrix * eigen_compliances[i])

        np.testing.assert_array_almost_equal(eigen_stiffnesses, np.sort(S.eig_stiffnesses))
        np.testing.assert_array_almost_equal(eigen_compliances, np.sort(C.eig_compliances))

    def test_to_from_Kelvin(self):
        C = S.inv()
        matrix = C.to_Kelvin()
        C2 = StiffnessTensor.from_Kelvin(matrix, symmetry=C.symmetry)
        np.testing.assert_array_almost_equal(C._matrix, C2._matrix)

    def test_ddot(self):
        # Check inverse
        C1 = StiffnessTensor.cubic(C11=22, C12=12, C44=44)
        C2 = StiffnessTensor.isotropic(E=210000, nu=0.3)
        np.testing.assert_almost_equal(C1.ddot(C1.inv()).full_tensor(), StiffnessTensor.identity().full_tensor())

        # Check product between tensor arrays
        m = 5
        rot_1d = orix_rot.random(m)
        C1_rot = C1 * rot_1d
        C2_rot = C1 * rot_1d

        C1C2_rot = C1_rot * C2
        assert C1C2_rot.shape == (m,)
        for i in range(m):
            np.testing.assert_array_almost_equal(C1C2_rot[i].full_tensor(), np.einsum('ijmn,nmkl->ijkl', C1_rot[i].full_tensor(), C2.full_tensor()))

        C1C2_rotrot = C1_rot * C2_rot
        assert C1C2_rot.shape == (m,)
        for i in range(m):
            np.testing.assert_array_almost_equal(C1C2_rotrot[i].full_tensor(), np.einsum('ijmn,nmkl->ijkl', C1_rot[i].full_tensor(),
                                                                        C2_rot[i].full_tensor()))

        C1C2_rotrot_cross = C1_rot.ddot(C2_rot, mode='cross')
        assert C1C2_rotrot_cross.shape == (m, m)
        for i in range(m):
            for j in range(m):
                np.testing.assert_array_almost_equal(C1C2_rotrot_cross[i,j].full_tensor(), np.einsum('ijmn,nmkl->ijkl', C1_rot[i].full_tensor(),
                                                                           C2_rot[j].full_tensor()))

    def test_full_tensor(self):
        C=S.inv()
        C_full = C.full_tensor()
        C2 = StiffnessTensor(C_full)
        np.testing.assert_array_almost_equal(C._matrix, C2._matrix)
        np.testing.assert_array_almost_equal(C.full_tensor(), C2.full_tensor())

    def test_voigt_reuss_axis(self):
        m,n = 5,6
        C=S.inv()
        orientations = orix_rot.random((m,n))
        C_rotated = C * orientations
        Cv_0 = C_rotated.Voigt_average(axis=0)
        Cr_0 = C_rotated.Reuss_average(axis=0)
        assert Cv_0.shape == (n,)
        assert Cr_0.shape == (n,)
        for i in range(n):
            assert Cv_0[i] == C_rotated[:, i].Voigt_average()
            assert Cr_0[i] == C_rotated[:, i].Reuss_average()
        Cv_1 = C_rotated.Voigt_average(axis=1)
        Cr_1 = C_rotated.Reuss_average(axis=1)
        assert Cv_1.shape == (m,)
        assert Cr_1.shape == (m,)
        for i in range(m):
            assert Cv_1[i] == C_rotated[i, :].Voigt_average()
            assert Cr_1[i] == C_rotated[i, :].Reuss_average()

        np.testing.assert_array_almost_equal(Cv_0.Voigt_average()._matrix, C_rotated.Voigt_average()._matrix)
        np.testing.assert_array_almost_equal(Cv_1.Voigt_average()._matrix, C_rotated.Voigt_average()._matrix)
        np.testing.assert_array_almost_equal(Cr_0.Reuss_average()._matrix, C_rotated.Reuss_average()._matrix)
        np.testing.assert_array_almost_equal(Cr_1.Reuss_average()._matrix, C_rotated.Reuss_average()._matrix)

    def test_invariants(self):
        C = S.inv()
        inv = C.invariants()
        C_rotated = C * rotations
        for i, inv_rotated_i in enumerate(C_rotated.invariants()):
            np.testing.assert_array_almost_equal(inv_rotated_i, inv[i])

if __name__ == '__main__':
    unittest.main()



import numpy as np
from Elasticipy.tensors.fourth_order import FourthOrderTensor
from Elasticipy.tensors.elasticity import StiffnessTensor
from scipy.integrate import trapezoid
from scipy.spatial.transform import Rotation

I = FourthOrderTensor.identity()

def gamma(C_macro_local, phi, theta, a1, a2, a3):
    s1 = np.sin(theta)*np.cos(phi) / a1
    s2 = np.sin(theta)*np.sin(phi) / a2
    s3 = np.cos(theta) / a3
    s = np.array([s1, s2, s3])
    D = np.einsum('lmnp,pqr,lqr->qrmn', C_macro_local.full_tensor(), s, s)
    Dinv = np.linalg.inv(D)
    return np.einsum('qrjk,iqr,lqr->qrijkl', Dinv, s, s) # The symmetrization is made afterward (see below)

def polarization_tensor(C_macro_local, a1, a2, a3, n_phi, n_theta):
    theta = np.linspace(0, np.pi, n_theta)
    phi = np.linspace(0, 2 * np.pi, n_phi)
    phi_grid, theta_grid = np.meshgrid(phi, theta, indexing='ij')
    g = gamma(C_macro_local, phi_grid, theta_grid, a1, a2, a3)
    gsin = (g.T*np.sin(theta_grid.T)).T
    a = trapezoid(gsin, phi, axis=0)
    b= trapezoid(a, theta, axis=0)/(4*np.pi)
    return FourthOrderTensor(b, force_minor_symmetry=True) # Symmetrization here (see above)

def localization_tensor(C_macro_local, C_incl, n_phi, n_theta, a1, a2, a3):
    E = polarization_tensor(C_macro_local, a1, a2, a3, n_phi, n_theta)
    delta = FourthOrderTensor(C_incl._matrix - C_macro_local._matrix)
    Ainv = E.ddot(delta) + I
    return Ainv.inv()

def Kroner_Eshelby(Ci, g, method='strain', max_iter=5, atol=1e-3, rtol=1e-3, display=False, n_phi=100, n_theta=100, particle_size=None):
    Ci_rotated = (Ci * g)
    C_macro = Ci_rotated.Hill_average()
    eigen_stiff = C_macro.eig_stiffnesses
    keep_on = True
    k = 0
    message = 'Maximum number of iterations is reached'
    m = len(g)
    A_local = FourthOrderTensor.zeros(m)
    if particle_size is None:
        a1 = a2 = a3 = 1
    else:
        a1, a2, a3 = particle_size
    while keep_on:
        eigen_stiff_old = eigen_stiff
        C_macro_local = C_macro * (g.inv())
        for i in range(m):
            A_local[i] = localization_tensor(C_macro_local[i], Ci, n_phi, n_theta, a1, a2, a3)
        A = A_local * g
        Q = Ci_rotated.ddot(A)
        if method=='stress':
            CiAi_mean = Q.mean()
            C_macro = StiffnessTensor.from_Kelvin(CiAi_mean._matrix, force_symmetries=True)
            err = A.mean() - FourthOrderTensor.identity()
        else:
            B = Q.ddot(C_macro.inv())
            R = Ci_rotated.inv().ddot(B)
            R_mean = R.mean()
            C_macro = StiffnessTensor.from_Kelvin(R_mean.inv()._matrix, force_symmetries=True)
            err = B.mean() - FourthOrderTensor.identity()

        # Stopping criteria
        eigen_stiff = C_macro.eig_stiffnesses
        abs_change = np.abs(eigen_stiff - eigen_stiff_old)
        rel_change = np.max(abs_change / eigen_stiff_old)
        max_abs_change = np.max(abs_change)
        k += 1
        if  max_abs_change < atol:
            keep_on = False
            message = 'Absolute change is below threshold value'
        if rel_change < rtol:
            keep_on = False
            message = 'Relative change is below threshold value'
        if k == max_iter:
            keep_on = False
        if display:
            err = np.max(np.abs(err._matrix))
            print('Iter #{}: abs. change={:0.5f}; rel. change={:0.5f}; error={:0.5f}'.format(k, max_abs_change, rel_change,err))
    return C_macro, message

C=StiffnessTensor.cubic(C11=108, C44=28.3, C12=62)
orientations = Rotation.random(1000)
Cmacro, msg = Kroner_Eshelby(C, orientations, method='stress', display=True)



import numpy as np
from Elasticipy.tensors.second_order import SymmetricSecondOrderTensor, rotation_to_matrix, is_orix_rotation, \
    SecondOrderTensor, ALPHABET
from scipy.spatial.transform import Rotation
from copy import deepcopy
from Elasticipy.tensors.mapping import KelvinMapping, VoigtMapping


def voigt_indices(i, j):
    """
    Translate the two-index notation to one-index notation

    Parameters
    ----------
    i : int or np.ndarray
        First index
    j : int or np.ndarray
        Second index

    Returns
    -------
    Index in the vector of length 6
    """
    voigt_mat = np.array([[0, 5, 4],
                          [5, 1, 3],
                          [4, 3, 2]])
    return voigt_mat[i, j]


def unvoigt_index(i):
    """
    Translate the one-index notation to two-index notation

    Parameters
    ----------
    i : int or np.ndarray
        Index to translate
    """
    inverse_voigt_mat = np.array([[0, 0],
                                  [1, 1],
                                  [2, 2],
                                  [1, 2],
                                  [0, 2],
                                  [0, 1]])
    return inverse_voigt_mat[i]

def rotate_tensor(full_tensor, r):
    """
    Rotate a (full) fourth-order tensor.

    Parameters
    ----------
    full_tensor : numpy.ndarray
        array of shape (3,3,3,3) or (...,3,3,3,3) containing all the components
    r : scipy.spatial.Rotation or orix.quaternion.Rotation
        Rotation, or set of rotations, to apply

    Returns
    -------
    numpy.ndarray
        Rotated tensor. If r is an array, the corresponding axes will be added as first axes in the result array.
    """
    rot_mat = rotation_to_matrix(r)
    str_ein = '...im,...jn,...ko,...lp,...mnop->...ijkl'
    return np.einsum(str_ein, rot_mat, rot_mat, rot_mat, rot_mat, full_tensor)

class FourthOrderTensor:
    """
    Template class for manipulating symmetric fourth-order tensors.

    Attributes
    ----------
    _matrix : np.ndarray
        (6,6) matrix gathering all the components of the tensor, using the Voigt notation.
    """
    tensor_name = '4th-order'

    def __init__(self, M, mapping=KelvinMapping(), check_minor_symmetry=True, force_minor_symmetry=False):
        """
        Construct of Fourth-order tensor with minor symmetry.

        Parameters
        ----------
        M : np.ndarray
            (6,6) matrix corresponding to the stiffness tensor, written using the Voigt notation, or array of shape
            (3,3,3,3).
        mapping : MappingConvention, optional
            Mapping convention to translate the (3,3,3,3) array to (6,6) matrix
        check_minor_symmetry : bool, optional
            If true (default), check that the input array have minor symmetries (see Notes). Only used if an array of
            shape (...,3,3,3,3) is passed.
        force_minor_symmetry :
            Ensure that the tensor displays minor symmetry.

        Notes
        -----
        The minor symmetry is defined so that:

        .. math::

            M_{ijkl}=M_{jikl}=M_{jilk}=M_{ijlk}

        """
        self.mapping=mapping
        M = np.asarray(M)
        if M.shape[-2:] == (6, 6):
            matrix = M
        elif M.shape[-4:] == (3, 3, 3, 3):
            Mijlk = np.swapaxes(M, -1, -2)
            Mjikl = np.swapaxes(M, -3, -4)
            Mjilk = np.swapaxes(Mjikl, -1, -2)
            if force_minor_symmetry:
                M = 0.25 * (M + Mijlk + Mjikl + Mjilk)
            elif check_minor_symmetry:
                symmetry = np.all(M == Mijlk) and np.all(M == Mjikl) and np.all(M == Mjilk)
                if not symmetry:
                    raise ValueError('The input array does not have minor symmetry')
            matrix = self._full_to_matrix(M)
        else:
            raise ValueError('The input matrix must of shape (...,6,6) or (...,3,3,3,3)')
        self._matrix = matrix
        for i in range(0, 6):
            for j in range(0, 6):
                def getter(obj, I=i, J=j):
                    return obj._matrix[...,I, J]

                getter.__doc__ = f"Returns the ({i + 1},{j + 1}) component of the {self.tensor_name} matrix."
                component_name = 'C{}{}'.format(i + 1, j + 1)
                setattr(self.__class__, component_name, property(getter))  # Dynamically create the property

    def __repr__(self):
        if (self.ndim == 0) or ((self.ndim==1) and self.shape[0]<5):
            msg = '{} tensor (in {} mapping):\n'.format(self.tensor_name, self.mapping.name)
            msg += self._matrix.__str__()
        else:
            msg = '{} tensor array of shape {}'.format(self.tensor_name, self.shape)
        return msg

    @property
    def shape(self):
        """
        Return the shape of the tensor array
        Returns
        -------
        tuple
            Shape of the tensor array
        """
        *shape, _, _ = self._matrix.shape
        return tuple(shape)

    def full_tensor(self):
        """
        Returns the full (unvoigted) tensor, as a [3, 3, 3, 3] array

        Returns
        -------
        np.ndarray
            Full tensor (4-index notation)
        """
        i, j, k, ell = np.indices((3, 3, 3, 3))
        ij = voigt_indices(i, j)
        kl = voigt_indices(k, ell)
        m = self._matrix[..., ij, kl] / self.mapping.matrix[ij, kl]
        return m

    def flatten(self):
        """
        Flatten the tensor

        If the tensor array is of shape (m,n,o...,r), the flattened array will be of shape (m*n*o*...*r,).

        Returns
        -------
        SymmetricFourthOrderTensor
            Flattened tensor
        """
        shape = self.shape
        if shape:
            t2 = deepcopy(self)
            p = (np.prod(self.shape), 6, 6)
            t2._matrix = self._matrix.reshape(p)
            return t2
        else:
            return self

    def _full_to_matrix(self, full_tensor):
        kl, ij = np.indices((6, 6))
        i, j = unvoigt_index(ij).T
        k, ell = unvoigt_index(kl).T
        return full_tensor[..., i, j, k, ell] * self.mapping.matrix[ij, kl]

    def rotate(self, rotation):
        """
        Apply a single rotation to a tensor, and return its component into the rotated frame.

        Parameters
        ----------
        rotation : Rotation or orix.quaternion.rotation.Rotation
            Rotation to apply

        Returns
        -------
        SymmetricFourthOrderTensor
            Rotated tensor
        """
        t2 = deepcopy(self)
        rotated_tensor = rotate_tensor(self.full_tensor(), rotation)
        t2._matrix = self._full_to_matrix(rotated_tensor)
        return t2

    @property
    def ndim(self):
        """
        Returns the dimensionality of the tensor (number of dimensions in the orientation array)

        Returns
        -------
        int
            Number of dimensions
        """
        shape = self.shape
        if shape:
            return len(shape)
        else:
            return 0

    def mean(self, axis=None):
        """
        Compute the mean value of the tensor T

        Parameters
        ----------
        axis : int or list of int or tuple of int, optional
            axis along which to compute the mean. If None, the mean is computed on the flattened tensor

        Returns
        -------
        numpy.ndarray
            If no axis is given, the result will be of shape (3,3,3,3).
            Otherwise, if T.ndim=m, and len(axis)=n, the returned value will be of shape (...,3,3,3,3), with ndim=m-n+4
        """
        t2 = deepcopy(self)
        if axis is None:
            axis = tuple([i for i in range(0,self.ndim)])
        t2._matrix = np.mean(self._matrix, axis=axis)
        return t2

    def __add__(self, other):
        if isinstance(other, np.ndarray):
            if other.shape == (6, 6):
                mat = self._matrix + other
            elif other.shape == (3, 3, 3, 3):
                mat = self._full_to_matrix(self.full_tensor() + other)
            else:
                raise ValueError('The input argument must be either a 6x6 matrix or a (3,3,3,3) array.')
        elif isinstance(other, FourthOrderTensor):
            if type(other) == type(self):
                mat = self.full_tensor() + other.full_tensor()
            else:
                raise ValueError('The two tensors to add must be of the same class.')
        else:
            raise ValueError('I don''t know how to add {} with {}.'.format(type(self), type(other)))
        return self.__class__(mat, mapping=self.mapping)

    def __sub__(self, other):
        if isinstance(other, FourthOrderTensor):
            return self.__add__(-other._matrix)
        else:
            return self.__add__(-other)

    def __neg__(self):
        t = deepcopy(self)
        t._matrix = -t._matrix
        return t

    def ddot(self, other, mode='pair'):
        """
        Perform tensor product contracted twice (":") between two fourth-order tensors

        Parameters
        ----------
        other : FourthOrderTensor or SecondOrderTensor
            Right-hand side of ":" symbol
        mode : str, optional
            If mode=="pair", the tensors must be broadcastable, and the tensor product are performed on the last axes.
            If mode=="cross", all cross-combinations are considered.

        Returns
        -------
        FourthOrderTensor or numpy.ndarray
         If both the tensors are 0D (no orientation), the return value will be of type SymmetricTensor
         Otherwise, the return value will be the full tensor, of shape (...,3,3,3,3).
        """
        if isinstance(other, FourthOrderTensor):
            if self.ndim == 0 and other.ndim == 0:
                return FourthOrderTensor(np.einsum('ijmn,nmkl->ijkl', self.full_tensor(), other.full_tensor()))
            else:
                if mode == 'pair':
                    ein_str = '...ijmn,...nmkl->...ijkl'
                else:
                    ndim_0 = self.ndim
                    ndim_1 = other.ndim
                    indices_0 = ALPHABET[:ndim_0]
                    indices_1 = ALPHABET[:ndim_1].upper()
                    indices_2 = indices_0 + indices_1
                    ein_str = indices_0 + 'wxXY,' + indices_1 + 'YXyz->' + indices_2 + 'wxyz'
                matrix = np.einsum(ein_str, self.full_tensor(), other.full_tensor())
                return FourthOrderTensor(matrix)
        elif isinstance(other, SecondOrderTensor):
            if self.ndim == 0 and other.ndim == 0:
                return SymmetricSecondOrderTensor(np.einsum('ijkl,kl->ij', self.full_tensor(), other.matrix))
            else:
                if mode == 'pair':
                    ein_str = '...ijkl,...kl->...ij'
                else:
                    ndim_0 = self.ndim
                    ndim_1 = other.ndim
                    indices_0 = ALPHABET[:ndim_0]
                    indices_1 = ALPHABET[:ndim_1].upper()
                    indices_2 = indices_0 + indices_1
                    ein_str = indices_0 + 'wxXY,' + indices_1 + 'XY->' + indices_2 + 'wx'
                matrix = np.einsum(ein_str, self.full_tensor(), other.matrix)
                return SecondOrderTensor(matrix)


    def __mul__(self, other):
        if isinstance(other, (FourthOrderTensor, SecondOrderTensor)):
            return self.ddot(other)
        elif isinstance(other, np.ndarray):
            shape = other.shape
            if other.shape == self.shape[-len(shape):]:
                matrix = self._matrix * other[...,np.newaxis, np.newaxis]
                return self.__class__(matrix)
            else:
                raise ValueError('The arrays to multiply could not be broadcasted with shapes {} and {}'.format(self.shape, other.shape[:-2]))
        elif isinstance(other, Rotation) or is_orix_rotation(other):
            return self.rotate(other)
        else:
            return self.__class__(self._matrix * other)

    def __truediv__(self, other):
        if isinstance(other, (SecondOrderTensor, FourthOrderTensor)):
            return self * other.inv()
        else:
            return self * (1 / other)


    def transpose_array(self):
        """
        Transpose the orientations of the tensor array

        Returns
        -------
        FourthOrderTensor
            The same tensor, but with transposed axes
        """
        ndim = self.ndim
        if ndim==0 or ndim==1:
            return self
        else:
            new_axes = tuple(range(ndim))[::-1] + (ndim, ndim + 1)
            transposed_matrix = self._matrix.transpose(new_axes)
            return self.__class__(transposed_matrix)

    def __rmul__(self, other):
        if isinstance(other, (Rotation, float, int, np.number)) or is_orix_rotation(other):
            return self * other
        else:
            raise NotImplementedError('A fourth order tensor can be left-multiplied by rotations or scalar only.')

    def __eq__(self, other):
        if isinstance(other, FourthOrderTensor):
            return np.all(self._matrix == other._matrix, axis=(-1, -2))
        elif isinstance(other, (float, int)) or (isinstance(other, np.ndarray) and other.shape[-2:] == (6, 6)):
            return np.all(self._matrix == other, axis=(-1, -2))
        else:
            raise NotImplementedError('The element to compare with must be a fourth-order tensor '
                                      'or an array of shape (6,6).')

    def __getitem__(self, item):
        if self.ndim:
            sub_mat= self._matrix[item]
            if sub_mat.shape[-2:] != (6,6):
                raise IndexError('Too many indices for tensor array: array is {}-dimensional, but {} were provided'.format(self.ndim, len(item)))
            else:
                return self.__class__(sub_mat)
        else:
            raise IndexError('A single tensor cannot be subindexed')

    def __setitem__(self, index, value):
        if isinstance(value, np.ndarray):
            if value.shape[-2:] == (6,6):
                self._matrix[index] = value
            elif value.shape[-4:] == (3,3,3,3):
                submatrix = self._full_to_matrix(value)
                self._matrix[index] = submatrix
            else:
                return ValueError('The R.h.s must be either of shape (...,6,6) or (...,3,3,3,3)')
        elif isinstance(value, FourthOrderTensor):
            self._matrix[index] = value._matrix / value.mapping.matrix * self.mapping.matrix
        else:
            raise NotImplementedError('The r.h.s must be either an ndarray or an object of class {}'.format(self.__class__))

    @classmethod
    def identity(cls, shape=(), return_full_tensor=False, mapping=KelvinMapping()):
        """
        Create a 4th-order identity tensor

        Parameters
        ----------
        shape : int or tuple, optional
            Shape of the tensor to create
        return_full_tensor : bool, optional
            If True, return the full tensor as a (3,3,3,3) or a (...,3,3,3,3) array. Otherwise, the tensor is returned
            as a SymmetricTensor object.
        mapping : str, optional
            Mapping convention to use. Must be either Kelvin or Voigt.

        Returns
        -------
        numpy.ndarray or SymmetricTensor
            Identity tensor
        """
        eye = np.eye(3)
        if isinstance(shape, int):
            shape = (shape,)
        if len(shape):
            for n in np.flip(shape):
                eye = np.repeat(eye[np.newaxis,...], n, axis=0)
        a = np.einsum('...ik,...jl->...ijkl', eye, eye)
        b = np.einsum('...il,...jk->...ijkl', eye, eye)
        full = 0.5*(a + b)
        if return_full_tensor:
            return full
        else:
            return cls(full, mapping=mapping)

    @classmethod
    def identity_spherical_part(cls, shape=(), return_full_tensor=False, mapping=KelvinMapping()):
        """
        Return the spherical part of the identity tensor

        Parameters
        ----------
        shape : tuple of int, optional
            Shape of the tensor to create
        return_full_tensor : bool, optional
            if true, the full tensor is returned as a (3,3,3,3) or a (...,3,3,3,3) array
        mapping : str, optional
            Mapping convention to use. Must be either Kelvin or Voigt.

        Returns
        -------
        FourthOrderTensor or SymmetricTensor
        """
        eye = np.eye(3)
        if isinstance(shape, int):
            shape = (shape,)
        if len(shape):
            for n in np.flip(shape):
                eye = np.repeat(eye[np.newaxis,...], n, axis=0)
        J = np.einsum('...ij,...kl->...ijkl',eye, eye)/3
        if return_full_tensor:
            return J
        else:
            return FourthOrderTensor(J, mapping=mapping)

    @classmethod
    def identity_deviatoric_part(cls, shape=(), return_full_tensor=False, mapping=KelvinMapping()):
        """
        Return the deviatoric part of the identity tensor

        Parameters
        ----------
        shape : tuple of int, optional
            Shape of the tensor to create
        return_full_tensor : bool, optional
            if true, the full tensor is returned as a (3,3,3,3) or a (...,3,3,3,3) array
        mapping : str, optional
            Mapping convention to use. Must be either Kelvin or Voigt.

        Returns
        -------
        FourthOrderTensor or SymmetricTensor
        """
        I = FourthOrderTensor.identity(shape, return_full_tensor, mapping)
        J = FourthOrderTensor.identity_spherical_part(shape, return_full_tensor, mapping)
        return I-J

    def spherical_part(self):
        """
        Return the spherical part of the tensor

        Returns
        -------
        FourthOrderTensor
            Spherical part of the tensor
        """
        I = self.identity_spherical_part(shape=self.shape)
        return I.ddot(self)

    def deviatoric_part(self):
        """
        Return the deviatoric part of the tensor

        Returns
        -------
        FourthOrderTensor
            Deviatoric part of the tensor
        """
        K = self.identity_deviatoric_part(shape=self.shape)
        return K.ddot(self)

    def inv(self):
        """
        Invert the tensor. The inverted tensors inherits the properties (if any)

        Returns
        -------
        FourthOrderTensor
            Inverse tensor
        """
        matrix_inv = np.linalg.inv(self._matrix)
        return self.__class__(matrix_inv, mapping=self.mapping.mapping_inverse)

    @classmethod
    def zeros(cls, shape=()):
        """
        Create a fourth-order tensor populated with zeros

        Parameters
        ----------
        shape : int or tuple, optional
            Shape of the tensor to create
        Returns
        -------
        FourthOrderTensor
        """
        if isinstance(shape, int):
            shape = (shape, 6, 6)
        else:
            shape = shape + (6,6)
        zeros = np.zeros(shape)
        return cls(zeros)

    def matrix(self, mapping_convention=None):
        matrix = self._matrix
        if mapping_convention is None:
            return matrix
        else:
            if isinstance(mapping_convention, str):
                if mapping_convention.lower() == 'voigt':
                    mapping_convention = VoigtMapping()
                elif mapping_convention.lower() == 'kelvin':
                    mapping_convention = KelvinMapping()
                else:
                    raise ValueError('Mapping convention must be either Kelvin or Voigt')
            return matrix / self.mapping._matrix * mapping_convention.matrix

class SymmetricFourthOrderTensor(FourthOrderTensor):
    tensor_name = 'Symmetric 4th-order'

    def __init__(self, M, check_symmetries=True, force_symmetries=False, **kwargs):
        """
        Construct a fully symmetric fourth-order tensor from a (...,6,6) or a (3,3,3,3) array.

        The input matrix must be symmetric, otherwise an error is thrown (except if check_symmetry==False, see below)

        Parameters
        ----------
        M : np.ndarray
            (6,6) matrix corresponding to the stiffness tensor, written using the Voigt notation, or array of shape
            (3,3,3,3).
        check_symmetries : bool, optional
            Whether to check or not that the tensor to built displays both major and minor symmetries (see Notes).
        force_symmetries : bool, optional
            If true, ensure that the tensor displays both minor and major symmetries.

        Notes
        -----
        The major symmetry is defined so that:

        .. math::

            M_{ijkl}=M_{klij}

        whereas the minor symmetry is:

        .. math::

            M_{ijkl}=M_{jikl}=M_{jilk}=M_{ijlk}
        """
        super().__init__(M, check_minor_symmetry=check_symmetries, force_minor_symmetry=force_symmetries, **kwargs)
        if force_symmetries:
            self._matrix = 0.5 * (self._matrix + self._matrix.swapaxes(-1, -2))
        elif check_symmetries and not np.all(np.isclose(self._matrix, self._matrix.swapaxes(-1, -2))):
            raise ValueError('The input matrix must be symmetric')

    def invariants(self, order='all'):
        """
        Compute the invariants of the tensor.

        Compute the linear or/and quadratic invariant of the fourth-order tensor (see notes)

        Parameters
        ----------
        order : str, optional
            If 'linear', only A1 and A2 are returned
            If 'quadratic', A1², A2², A1*A2, B1, B2, B3, B4 and B5 are returned
            If 'all' (default), A1, A2, A1², A2², B1, B2, B3, B4 and B5 are returned

        Returns
        -------
        tuple
            invariants of the given order (see above)

        Notes
        -----
        The nomenclature of the invariants follows that of [4]_. The linear invariants are:

        .. math::

            A_1=C_{ijij}

            A_2=C_{iijj}

        whereas the quadratic invariants are:

        .. math::

            B_1 = C_{ijkl}C_{ijkl}

            B_2 = C_{iikl}C_{jjkl}

            B_3 = C_{iikl}C_{jkjl}

            B_4 = C_{kiil}C_{kjjl}

            B_5 = C_{ijkl}C_{ikjl}

        References
        ----------
        .. [4] Norris, A. N. (22 May 2007). "Quadratic invariants of elastic moduli". The Quarterly Journal of Mechanics
         and Applied Mathematics. 60 (3): 367–389. doi:10.1093/qjmam/hbm007
        """
        t = self.full_tensor()
        order = order.lower()
        A1 = np.einsum('...ijij->...',t)
        A2 = np.einsum('...iijj->...',t)
        lin_inv = (A1, A2)
        if order == 'linear':
            return lin_inv
        B1 = np.einsum('...ijkl,...ijkl->...',t, t)
        B2 = np.einsum('...iikl,...jjkl->...', t, t)
        B3 = np.einsum('...iikl,...jkjl->...', t, t)
        B4 = np.einsum('...kiil,...kjjl->...', t, t)
        B5 = np.einsum('...ijkl,...ikjl->...', t, t)
        quad_inv = (A1**2, A2**2, A1*A2, B1, B2, B3, B4, B5)
        if order == 'quadratic':
            return quad_inv
        else:
            return lin_inv + quad_inv



import warnings
from Elasticipy.tensors.thermal_expansion import ThermalExpansionTensor as NewThermalExpansionTensor

warnings.warn(
    "The module 'Elasticipy.ThermalExpansion' is deprecated and will be removed in a future release. "
    "Please use 'Elasticipy.tensors.thermal_expansion' instead.",
    DeprecationWarning,
    stacklevel=2
)

class ThermalExpansionTensor(NewThermalExpansionTensor):
    pass


import unittest
import numpy as np
from Elasticipy.plasticity import JohnsonCook
from Elasticipy.plasticity import TrescaPlasticity, VonMisesPlasticity, DruckerPrager
from pytest import approx

from Elasticipy.tensors.stress_strain import StressTensor, StrainTensor

A, B, C = 792, 510, 0.014
m, n = 1.03, 0.26
eps_dot_ref = 1
T0, Tm = 25, 1500
JC    = JohnsonCook(A=A, B=B, n=n)
JC_rd = JohnsonCook(A=A, B=B, n=n, C=C, eps_dot_ref=eps_dot_ref)
JC_td = JohnsonCook(A=A, B=B, n=n, m=1.03, T0=T0, Tm=Tm)
JC_rtd= JohnsonCook(A=A, B=B, n=n, C=C, eps_dot_ref=eps_dot_ref, m=m, T0=T0, Tm=Tm)
K = 3 / 2 * 1 / 3 ** 0.5
JC_tresca = JohnsonCook(A=A, B=B, n=n, criterion='Tresca')


class TestJohnsonCook(unittest.TestCase):
    def test_JC_string(self):
        assert JC.__repr__() == ('Johnson-Cook plasticity model\n'
                                 ' type: Isotropic\n'
                                 ' criterion: von Mises\n'
                                 ' current strain: 0.0')
        JC.apply_strain(0.1)
        assert JC.__repr__() == ('Johnson-Cook plasticity model\n'
                                 ' type: Isotropic\n'
                                 ' criterion: von Mises\n'
                                 ' current strain: 0.1')
        JC.reset_strain()
        assert JC.__repr__() == ('Johnson-Cook plasticity model\n'
                                 ' type: Isotropic\n'
                                 ' criterion: von Mises\n'
                                 ' current strain: 0.0')
        assert JC_tresca.__repr__() == ('Johnson-Cook plasticity model\n'
                                        ' type: Isotropic\n'
                                        ' criterion: Tresca\n'
                                        ' current strain: 0.0')

    def test_yield_stress(self):
        assert JC.flow_stress(0) == A
        assert JC_rd.flow_stress(0, eps_dot=eps_dot_ref) == A
        assert JC_td.flow_stress(0, T=T0) == A

    def test_rate_dependence(self):
        assert JC_rd.flow_stress(0.1, eps_dot=2) == JC_rd.flow_stress(0.1, eps_dot=1) * (1 + C*np.log(2))
        for model in (JC, JC_td):   # Check that an error is thrown if the model is not rate-dependent
            with self.assertRaises(ValueError) as context:
                _ = model.flow_stress(0.1, eps_dot=2)
            self.assertEqual(str(context.exception), 'C and eps_dot_ref must be defined for using a rate-dependent model')

    def test_temperature_dependence(self):
        assert JC_td.flow_stress(0.1, T=T0) == JC.flow_stress(0.1)
        assert JC_td.flow_stress(0.1, T=Tm) == 0.0
        for model in (JC, JC_rd):   # Check that an error is thrown if the model is not temperature-dependent
            with self.assertRaises(ValueError) as context:
                _ = model.flow_stress(0.1, T=T0)
            self.assertEqual(str(context.exception), 'T0, Tm and m must be defined for using a temperature-dependent model')

    def test_compute_strain_increment(self):
        JC2 = JohnsonCook(A=A, B=B, n=n)
        strain0 = 0.1

        # Test temperature-independent model
        stress = JC2.flow_stress(strain0)
        strain1 = JC2.compute_strain_increment(stress)
        assert strain1 == approx(strain0)

        # Now try with a full tensor
        JC.reset_strain()
        assert JC.plastic_strain == 0.0
        stress = StressTensor.tensile([1,0,0], A+10)
        strain = JC.compute_strain_increment(stress)
        normalized_strain = strain / strain.eq_strain()
        np.testing.assert_array_almost_equal(normalized_strain.matrix, np.diag([1, -0.5, -0.5]))

        # Test temperature-dependent model
        stress = JC_td.flow_stress(strain0, T=500)
        strain2 = JC_td.compute_strain_increment(stress, T=500)
        assert strain2 == approx(strain0)

        # What if we use try to use the temperature on temperature-independent model
        with self.assertRaises(ValueError) as context:
            _ = JC.flow_stress(0.1, T=T0)
        self.assertEqual(str(context.exception), 'T0, Tm and m must be defined for using a temperature-dependent model')

        # Check that if stress < A, the strain is zero
        assert JC.compute_strain_increment(A) == 0.0

        # Check that if the temperature is larger than Tm, the strain is infinite
        assert JC_td.compute_strain_increment(0, T=Tm) == np.inf


    def test_normality_J2(self):
        tensile_stress = StressTensor.tensile([1,0,0], 1)
        normal = VonMisesPlasticity.normal(tensile_stress)
        assert normal == np.diag([1., -0.5, -0.5])

        shear_stress = StressTensor.shear([1, 0, 0], [0, 1, 0], 1)
        normal = VonMisesPlasticity.normal(shear_stress)
        normal_th = K * np.array([[0, 1, 0],
                                  [1, 0, 0],
                                  [0, 0, 0]])
        np.testing.assert_array_almost_equal(normal.matrix, normal_th)

    def test_normality_Tresca(self):
        biaxial = (StressTensor.tensile([1,0,0],[0, 1, 1, 1, 1, 1, 0]) +
                   StressTensor.tensile([0,1,0],[-1, -1, -0.5, 0, 0.5, 1, 1]))
        n = TrescaPlasticity.normal(biaxial)
        assert n[0] == VonMisesPlasticity.normal(biaxial[0])
        assert n[2] == K * np.diag([1, -1, 0])
        assert n[2] == K * np.diag([1, -1, 0])
        assert n[3] == VonMisesPlasticity.normal(biaxial[3])
        assert n[4] == K * np.diag([1, 0, -1])
        assert n[5] == VonMisesPlasticity.normal(biaxial[5])
        assert n[6] == VonMisesPlasticity.normal(biaxial[6])

        # Check that the magnitude of the normal is 1
        np.testing.assert_array_equal(n.eq_strain(), np.ones(biaxial.shape))
        triaxial = StressTensor(np.diag([1,2,4]))
        n = TrescaPlasticity.normal(triaxial)
        assert n == K * np.diag([-1, 0, 1])
        assert n.eq_strain() == 1.0


    def test_apply_strain(self):
        strain = StrainTensor.tensile([1,0,0], 1) + StrainTensor.tensile([0, 1, 0], -0.5) + StrainTensor.tensile([0, 0, 1], -0.5)
        JC.apply_strain(0.0)    # Try with float
        assert JC.plastic_strain == 0.0
        JC.apply_strain(strain) # Try with StrainTensor
        assert JC.plastic_strain == 1.0
        JC.apply_strain(-1.0)
        assert JC.plastic_strain == 2

    def test_Tresca_plasticity(self):
        JC_tresca = JohnsonCook(A=A, B=B, n=n, criterion='Tresca')
        stress = StressTensor.shear([1,0,0],[0,1,0], 1000)
        strain = JC_tresca.compute_strain_increment(stress)
        eq_stress_tr = JC_tresca.flow_stress(strain.eq_strain())
        assert eq_stress_tr == approx(stress.Tresca())

    def test_DruckerPrager(self):
        dp = DruckerPrager(0.2)
        JC_pg = JohnsonCook(A=A, B=B, n=n, criterion=dp)
        shear_stress = StressTensor.shear([1, 0, 0], [0, 1, 0], 1000)
        tens_shear_stress = shear_stress + StressTensor.eye() * 100
        comp_shear_stress = shear_stress - StressTensor.eye() * 100
        strain_0 = JC_pg.compute_strain_increment(shear_stress, apply_strain=False)
        strain_p = JC_pg.compute_strain_increment(tens_shear_stress, apply_strain=False)
        strain_m = JC_pg.compute_strain_increment(comp_shear_stress, apply_strain=False)
        assert strain_m.eq_strain() < strain_0.eq_strain() < strain_p.eq_strain()

        # Now investigate the special case alpha=0 (== von Mises)
        JC_pg0 = JohnsonCook(A=A, B=B, n=n, criterion=DruckerPrager(0.))
        JC.reset_strain()
        strain_0 = JC_pg0.compute_strain_increment(shear_stress, apply_strain=False)
        strain_p = JC_pg0.compute_strain_increment(tens_shear_stress, apply_strain=False)
        strain_m = JC_pg0.compute_strain_increment(comp_shear_stress, apply_strain=False)
        strain_vm = JC.compute_strain_increment(tens_shear_stress, apply_strain=False)
        assert strain_0.eq_strain() == approx(strain_p.eq_strain())
        assert strain_0.eq_strain() == approx(strain_m.eq_strain())
        assert strain_0.eq_strain() == approx(strain_vm.eq_strain())



if __name__ == '__main__':
    unittest.main()



from Elasticipy.tensors.second_order import SecondOrderTensor, SymmetricSecondOrderTensor
from Elasticipy.tensors.stress_strain import StressTensor
import pandas as pd
import numpy as np

def _rebuild_tensor(F11, F22, F33, F12, F13, F21, F23, F31, F32):
    return np.array([[F11, F12, F13], [F21, F22, F23], [F31, F32, F33]]).transpose((2,0,1))

def from_quadrature_file(file, returns='stress'):
    """
    Read data from quadrature output file generated by PRISMS plasticity.

    These files, usually named QuadratureOutputsXXX.csv (where XXX denotes the time increment), contain large amount of
    data, such as gradient components, stress values, grain ID etc.

    Parameters
    ----------
    file : str
        Path to quadrature file
    returns : str or list of str or tuple of str
        name(s) of requested field(s). They can be:
          - grain ID
          - phase ID
          - det(J)
          - twin
          - coordinates
          - orientation
          - elastic gradient
          - plastic gradient
          - stress
    Returns
    -------
    ndarray or SecondOrderTensor or StressTensor
        The number of returned values depends on the requested fields.
    """
    data=pd.read_csv(file, header=None, dtype=float, usecols=range(0,37))
    grainID, phaseID, detJ, twin, x, y, z, rot1, rot2, rot3, *other = data.T.to_numpy()
    Fe11, Fe22, Fe33, Fe12, Fe13, Fe21, Fe23, Fe31, Fe32, *FpStress = other
    Fp11, Fp22, Fp33, Fp12, Fp13, Fp21, Fp23, Fp31, Fp32, *stress_compo = FpStress
    if isinstance(returns, str):
        returns_list = (returns,)
    else:
        returns_list = returns
    returned_values = []
    for r in returns_list:
        rl = r.lower()
        if (rl == 'grain id') or (rl == 'grainid'):
            returned_values.append(grainID)
        elif (rl == 'phase id') or (rl == 'phaseid'):
            returned_values.append(phaseID)
        elif rl == 'det(j)':
            returned_values.append(detJ)
        elif rl == 'twin':
            returned_values.append(twin)
        elif rl == 'coordinates':
            returned_values.append(np.array([x, y, z]).T)
        elif rl == 'orientation':
            returned_values.append(np.array([rot1, rot2, rot3]).T)
        elif rl == 'elastic gradient':
            Fe = _rebuild_tensor(Fe11, Fe22, Fe33, Fe12, Fe13, Fe21, Fe23, Fe31, Fe32)
            returned_values.append(SecondOrderTensor(Fe))
        elif rl == 'plastic gradient':
            Fp = _rebuild_tensor(Fp11, Fp22, Fp33, Fp12, Fp13, Fp21, Fp23, Fp31, Fp32)
            returned_values.append(SecondOrderTensor(Fp))
        elif rl == 'stress':
            stress = _rebuild_tensor(*stress_compo)
            returned_values.append(StressTensor(stress))
        else:
            raise ValueError('Unknown return type')
    if isinstance(returns, str):
        return returned_values[0]
    else:
        return tuple(returned_values)

def from_stressstrain_file(file):
    """
    Read data from stress-strain file generated by PRISMS plasticity.

    This file is usually named stressstrain.txt

    Parameters
    ----------
    file : str
        Path to stress-strain file

    Returns
    -------
    SymmetricSecondOrderTensor
        Average Green-Lagrange strain tensor
    StressTensor
        Average stress tensor
    """
    data = pd.read_csv(file, sep='\t')
    Exx, Exy, Exz = data['Exx'], data['Exy'], data['Exz']
    Eyy, Eyz = data['Eyy'], data['Eyz']
    Ezz = data['Ezz']
    E = np.array([[Exx, Exy, Exz], [Exy, Eyy, Eyz], [Exz, Eyz, Ezz]]).transpose((2,0,1))
    Txx, Txy, Txz = data['Txx'], data['Txy'], data['Txz']
    Tyy, Tyz = data['Tyy'], data['Tyz']
    Tzz = data['Tzz']
    T = np.array([[Txx, Txy, Txz], [Txy, Tyy, Tyz], [Txz, Tyz, Tzz]]).transpose((2, 0, 1))
    return SymmetricSecondOrderTensor(E), StressTensor(T)

